import asyncio
import itertools
import random

import nest_asyncio
import numpy as np
from shapely.affinity import translate
from shapely.geometry import LineString, Point
from shapely.ops import nearest_points

from kmService import KmLintMeasure
from kmService.km_models import KmValueObject
from kmService.km_responses import (
    KmResponse,
    ProjectionInputResponse,
    ProjectionResponse,
)
from kmService.km_service_builder import KmServiceBuilder
from kmService.uitils.shapely_geojson import (
    GeoJsonFeature,
    GeoJsonFeatureCollection,
    dumps,
)
from kmService.uitils.shapely_helpers import ShapelyTransform

nest_asyncio.apply()


class KmService:
    """
    A service class for calculating kilometers and related features.
    """

    def __init__(self):
        self._value_objects_dict: dict[str, KmValueObject] = {}

    @classmethod
    async def factory(cls):
        """
        Factory method to create an instance of KmService.

        Returns:
            KmService: An instance of KmService.
        """
        self = cls()
        km_service = await KmServiceBuilder.factory(
            "https://maps.prorail.nl/arcgis/rest/services/Referentiesysteem/FeatureServer"
        )
        self._value_objects_dict = km_service.value_objects
        return self

    async def _get_raaien_to_measure(self, point: Point) -> KmResponse:
        # in some cases we can have more than  one set of raaien (flyover situation)
        async def process(point_, item_):
            if point_.within(item_.geometry):
                init_response = KmLintMeasure.from_value_object(item_, point)

                # find closest 2 raaien pick lowest
                distance_raai_to_point = dict(
                    sorted(
                        {
                            raai.geometry_corrected.distance(point): raai
                            for raai in item_.km_raaien
                        }.items()
                    )
                )
                closest_two_raaien = [
                    distance_raai_to_point[k] for k in list(distance_raai_to_point)[:2]
                ]
                if closest_two_raaien[0].hectometer < closest_two_raaien[1].hectometer:
                    init_response.raai = closest_two_raaien[0].geometry_corrected
                    init_response.hm = closest_two_raaien[0].hectometer
                    return init_response
                else:
                    init_response.raai = closest_two_raaien[1].geometry_corrected
                    init_response.hm = closest_two_raaien[1].hectometer
                    return init_response

        tasks = []
        tasks.extend(
            [process(point, item) for item in self._value_objects_dict.values()]
        )
        result = await asyncio.gather(*tasks)
        return KmResponse(point, [item for item in result if item is not None])

    async def _get_km_async(self, x: float, y: float) -> KmResponse:
        point = Point(x, y)
        responses = await self._get_raaien_to_measure(point)
        for response in responses.km_measures:
            projection_point = response.raai.interpolate(response.raai.project(point))
            distance = projection_point.distance(point)
            response.distance = distance
        return responses

    def get_km(self, x: float, y: float) -> KmResponse:
        """
        Retrieves kilometer responses based on the provided coordinates.

        Args:
            x: The x-coordinate.
            y: The y-coordinate.

        Returns:
            A list of kilometer responses.
        """
        response = asyncio.run(self._get_km_async(x, y))
        return response

    async def _get_km_batch_async(self, point_list) -> list[KmResponse]:
        tasks = [self._get_km_async(point[0], point[1]) for point in point_list]
        results = await asyncio.gather(*tasks)
        return [result for result in results]

    def get_km_batch(self, point_list: list[list[int | float]]) -> list[KmResponse]:
        """
        Retrieves KM responses synchronously for a batch of points.

        Args:
            point_list: A list of lists containing the coordinates of points.

        Returns:
            A list of lists containing KM responses for each point.
        """
        response = asyncio.run(self._get_km_batch_async(point_list))
        return response

    def get_lint_names(self) -> list[str]:
        """
        Get all lint names.

        Returns:
            A list of unique lint names.
        """
        return list(
            set(
                [
                    value.km_vlak.km_lint_naam
                    for key, value in self._value_objects_dict.items()
                ]
            )
        )

    def project_point(
        self, lint_name: str, hm: float, m: int, line_geometry: LineString
    ) -> ProjectionResponse:
        """

        Get a projected point on the line.

        !!! warning "This feature is experimental and may not be suitable for production use."
            ***Api and response models will change!***
            Ensure to thoroughly test this feature with your data before deploying it.

        !!! info "Way of working"

            1. Get all raaien associated with the input lint.
            2. Find the closest raaien: the one before and after the specified kilometer measure.
            3. Calculate reference and target raai: based on the hm found before and after.
                a. if no next we use raai before and negative distance to project.
            4. Find nearest points of the reference line and the target line.
            5. Calculate movement vector from the nearest points.
            6. Move the reference line towards the target line.
            j. Calculate projected point: intersection point of the input line geometry with the moved reference line.

        Args:
            lint_name: The name of the lint.
            hm: Hectometer value to measure from.
            m: Distance in meters to measure.
            line_geometry: The geometry of the line, should match the projected kilometering line.

        Returns:
            ProjectionResponse: Response containing projected point information.
        """

        lint_raaien = list(
            itertools.chain.from_iterable(
                [
                    value.km_raaien
                    for key, value in self._value_objects_dict.items()
                    if key.endswith(lint_name)
                ]
            )
        )

        # Initialize variables to hold the objects before and after the input value
        object_before = None
        object_after = None

        # Initialize variables to hold the minimum differences
        min_difference_before = float("inf")
        min_difference_after = float("inf")

        # get raai of interest
        raai_to_project_from = [_ for _ in lint_raaien if _.hectometer == hm][0]

        # debug geojson
        reference_line = raai_to_project_from.geometry_corrected

        # Iterate through the list of objects get raai before and after...
        for obj in lint_raaien:
            difference = abs(obj.hectometer - hm)
            if obj.hectometer < hm and difference < min_difference_before:
                min_difference_before = difference
                object_before = obj
            elif obj.hectometer > hm and difference < min_difference_after:
                min_difference_after = difference
                object_after = obj

        # Determine the line to project towards
        if object_after:
            target_line = object_after.geometry_corrected
        elif object_before:
            # use before and make movement negative so offset point the right direction
            target_line = object_before.geometry_corrected
            m *= -1
        else:
            raise NotImplementedError("can not handle single raai in lint.")

        # Find the nearest points on both lines
        nearest_point_reference, nearest_point_target = nearest_points(
            reference_line, target_line
        )

        # Calculate the vector from the reference line's starting point to the nearest point on the target line
        dx_nearest = (
            nearest_point_target.coords[0][0] - nearest_point_reference.coords[0][0]
        )
        dy_nearest = (
            nearest_point_target.coords[0][1] - nearest_point_reference.coords[0][1]
        )
        unit_direction_nearest = np.array([dx_nearest, dy_nearest]) / np.linalg.norm(
            [dx_nearest, dy_nearest]
        )

        # Define the distance to move the reference line
        distance = (
            m - 0.001 if m < 0 else m + 0.001 if m > 0 else m
        )  # Adjust it a bit so it is not just under.

        # Move the reference line towards the target line
        movement_vector_nearest = unit_direction_nearest * distance
        moved_reference_line = translate(
            reference_line, movement_vector_nearest[0], movement_vector_nearest[1]
        )

        projected_point = line_geometry.intersection(moved_reference_line)
        if projected_point.is_empty:
            raise ValueError("input line geometry does not intersect the km_line")  # noqa TRY003
        return ProjectionResponse(
            input_data=ProjectionInputResponse(
                input_line=line_geometry,
                input_lint_name=lint_name,
                input_hm=hm,
                input_distance=m,
            ),
            km_line=moved_reference_line,
            km_point=projected_point,
            reference_line=reference_line,
        )

    def map_geojson_string(
        self, lint_names: list[str] | None = None, file_name: str | None = None
    ) -> str:
        """
        Generates GeoJSON output based on the data stored in KmService that can be used to plot a coverage map.

        Args:
            lint_names: A list of lint names to filter the data. If None, all data will be included.
            file_name: The name of the output file. If None, the GeoJSON string will be returned instead of writing to a file.

        Returns:
            str: A GeoJSON string representing the data.
        """

        def add_feature():
            features.append(
                GeoJsonFeature(
                    [ShapelyTransform.rd_to_wgs(value.geometry)],
                    {
                        "color": picked_color,
                        "fill-color": "transparent",
                        "fill-opacity": 0.5,
                    },
                )
            )

            for item in value.matched:
                # plot red if not matched
                if not item[1]:
                    features.append(
                        GeoJsonFeature(
                            [ShapelyTransform.rd_to_wgs(item[0].geometry)],
                            {
                                "color": "red",
                                "radius": "5px",
                                "type": "mismatched_hm_point",
                                "hm": item[0].hectometer,
                            },
                        )
                    )  # | item[0].__dict__))
                else:
                    features.append(
                        GeoJsonFeature(
                            [ShapelyTransform.rd_to_wgs(item[0].geometry)],
                            {
                                "color": picked_color,
                                "radius": "1px",
                                "type": "matched_hm_point",
                                "hm": item[0].hectometer,
                            },
                        )
                    )
                    features.append(
                        GeoJsonFeature(
                            [ShapelyTransform.rd_to_wgs(item[1].geometry)],
                            {
                                "color": "#c9c9c9",
                                "type": "matched_raai",
                                "hm": item[1].hectometer,
                            },
                        )
                    )
                    # all so plot corrected so we can see the movement
                    features.append(
                        GeoJsonFeature(
                            [ShapelyTransform.rd_to_wgs(item[1].geometry_corrected)],
                            {
                                "color": picked_color,
                                "type": "moved_raai",
                                "hm": item[1].hectometer,
                            },
                        )
                    )

        colors = [
            "#1F77B4",
            "#AEC7E8",
            "#FF7F0E",
            "#FFBB78",
            "#2CA02C",
            "#98DF8A",
            "#FF9896",
            "#9467BD",
            "#C5B0D5",
            "#8C564B",
            "#C49C94",
            "#E377C2",
            "#F7B6D2",
            "#7F7F7F",
            "#C7C7C7",
            "#BCBD22",
            "#DBDB8D",
            "#17BECF",
            "#9EDAE5",
        ]

        features: list[GeoJsonFeature] = []
        for key, value in self._value_objects_dict.items():
            picked_color = random.choice(colors)
            if lint_names is not None:
                if value.km_vlak.km_lint_naam in lint_names:
                    add_feature()
            else:
                add_feature()

        geojson_string = dumps(GeoJsonFeatureCollection(features))
        if file_name is not None:
            with open(file_name, "w") as text_file:
                text_file.write(geojson_string)
        return geojson_string


def get_km_service() -> KmService:
    """
    Retrieves an instance of KmService using asyncio.

    Returns:
        KmService: An instance of KmService.
    """
    return asyncio.run(KmService.factory())



from __future__ import annotations

import json
from typing import IO, cast

from shapely.geometry import LineString, Point, Polygon, mapping
from shapely.geometry.base import BaseGeometry

# todo make own pipy module and make dependency


class GeoJsonFeature:
    """
    A geojson feature build from one or more shapely geometries and a set of properties as a dictionary.

    It's possible to use multi geometry types, but most applications wont support geojson multigeometry types, same as nested property dictionaries.

    Args:
        geometry_list: list of shapley geometry
        properties: dictionary of properties of the feature

    """

    def __init__(
        self, geometry_list: list[BaseGeometry], properties: dict | None = None
    ) -> None:
        self.geometry_list: list[BaseGeometry] = geometry_list
        self.properties: dict | None = properties

    @property
    def geometry_list(self):
        return self.__geometry_list

    @geometry_list.setter
    def geometry_list(self, geometry_list: list[BaseGeometry]):
        for item in geometry_list:
            if not isinstance(item, BaseGeometry):
                raise TypeError("geometry entry must be a shapely geometry.")  # noqa RY003
        self.__geometry_list = geometry_list

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, properties):
        if properties is None:
            self._properties = {}
        elif isinstance(properties, dict):
            self._properties = properties
        else:
            raise ValueError("properties must be a dict.")  # noqa RY003

    def _get_geo_interface(self):
        geometries = []

        if len(self.geometry_list) == 0:
            geometries = None

        elif len(self.geometry_list) > 1:
            for item in self.geometry_list:
                if geometries is not None:
                    geometries.append(item.__geo_interface__)
        else:
            if self.geometry_list[0].is_empty:
                geometries = None
            else:
                geometries = self.geometry_list[0].__geo_interface__

        return geometries

    @property
    def __geo_interface__(self):
        geometry = self._get_geo_interface()
        geo_json_type = "Feature"

        if len(self.__geometry_list) == 0:
            geometry = {"geometry": None}
        elif len(self.__geometry_list) == 1:
            geometry = {"geometry": geometry}
        elif len(self.__geometry_list) > 1:
            geometry = {
                "geometry": {"type": "GeometryCollection", "geometries": geometry}
            }

        return {
            "type": geo_json_type,
            "properties": self.properties,
        } | geometry

    def __eq__(self, other):
        return self.__geo_interface__ == other.__geo_interface__

    def __repr__(self):
        return f"<Feature {str(self.__geo_interface__)[1:-1]}>"

    def as_dict(self):
        return self.__geo_interface__


class GeoJsonFeatureCollection:
    """
    GeoJson FeatureCollection stores geojson Features.

    Args:
        geojson_features (list[GeoJsonFeature]):

    """

    def __init__(self, geojson_features: list[GeoJsonFeature | BaseGeometry]):
        # todo: make crs init parameter
        # todo: validate if features are in crs range
        self.features: list[GeoJsonFeature] = geojson_features
        self.crs: dict = {
            "crs": {
                "type": "name",
                "properties": {
                    "name": "urn:ogc:def:crs:EPSG::4326"
                    # RD: http://www.opengis.net/def/crs/EPSG/0/28992 # RD/NAP: EPSG:7415 # CRS84: 4326
                },
            }
        }

    @property
    def features(self) -> list[GeoJsonFeature]:
        """Get all features."""
        return self._features

    @features.setter
    def features(self, objects: list[GeoJsonFeature | BaseGeometry]):
        all_are_features = all(
            isinstance(feature, GeoJsonFeature) for feature in objects
        )
        if all_are_features:
            self._features: list[GeoJsonFeature] = objects
        else:
            for item in objects:
                if not isinstance(item, BaseGeometry):
                    raise TypeError(  # noqa RY003
                        "features can be either a Feature or shapely geometry."  # noqa RY003
                    )  # noqa RY003
            self._features = [
                GeoJsonFeature(cast(list[BaseGeometry], [geometry]))
                for geometry in objects
            ]

    def __iter__(self):
        return iter(self.features)

    def geometries_iterator(self):
        for feature in self.features:
            if len(feature.geometry_list) > 1:
                yield from feature.geometry_list
            else:
                yield feature.geometry_list[0]

    @property
    def __geo_interface__(self):
        return {
            "crs": self.crs["crs"],
            "type": "FeatureCollection",
            "features": [feature.__geo_interface__ for feature in self.features],
        }

    def __eq__(self, other):
        return self.__geo_interface__ == other.__geo_interface__

    def __repr__(self):
        return f"<FeatureCollection {str(self.__geo_interface__)[1:-1]}>"

    def as_dict(self):
        return self.__geo_interface__

    def as_string(self) -> str:
        """Returns geojson string."""
        return dumps(self)


class ShapleyGeojsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Point | LineString | Polygon):
            return obj.wkt
        return json.JSONEncoder.default(self, obj)


def dump(
    feature_collection: GeoJsonFeatureCollection, file_path: IO[str], *args, **kwargs
) -> None:
    """
    Dump FeatureCollection to file.

    Args:
        feature_collection: FeatureCollection
        file_path: output file path
    """
    json.dump(
        mapping(feature_collection),
        file_path,
        cls=ShapleyGeojsonEncoder,
        *args,
        **kwargs,
    )


def dumps(feature_collection: GeoJsonFeatureCollection, *args, **kwargs) -> str:
    """
    Dump FeatureCollection to string.

    Args:
        feature_collection (GeoJsonFeatureCollection): FeatureCollection

    Returns:
        (str)): geojson string representation of the FeatureCollection.
    """
    return json.dumps(
        mapping(feature_collection), cls=ShapleyGeojsonEncoder, *args, **kwargs
    )



from dataclasses import dataclass, field

from shapely.geometry import LineString, Point, Polygon

from kmService.uitils.shapleyTools import extend_line


@dataclass
class KmVlak:
    object_id: int
    km_lint_naam: str
    km_lint_omschrijving: str
    geometry: Polygon

    def xy_within(self, x: float, y: float):
        return Point(x, y).within(self.geometry)


@dataclass
class Raai:
    geometry_corrected: LineString = None


@dataclass
class KmRaai:
    object_id: int
    geo_code: str
    sub_code: str
    _hectometer_str: str
    geometry: LineString
    geometry_extended: LineString | None = None
    geometry_corrected: LineString | None = None
    hectometer: float = 0

    def __post_init__(self):
        extend_offset = 500
        self.hectometer = float(self._hectometer_str.replace(",", "."))
        self.geometry_extended = extend_line(self.geometry, extend_offset)

    def __str__(self):
        return (
            f"geocode:{self.geo_code} sub_geocode={self.sub_code} hm={self.hectometer}"
        )


# todo: make config file, and parse to objects so we can add objects to value_objects_dict
#  use case: missing raaien, and new tracé, future feature is to use to build raaien...
@dataclass
class CustomKmRaai:
    lint_name: str
    lint_afkorting: str
    hectometer: float
    geocode: str
    sub_code: str
    geometry: LineString
    object_id: str = field(init=False)

    def from_dict(self):
        pass


@dataclass
class KmPunt:
    object_id: int
    geo_code: str
    sub_code: str
    hectometer: float
    km_lint: str
    km_lint_omschrijving: str
    geometry: LineString

    def __str__(self):
        return f"km_lint={self.km_lint} geocode={self.geo_code} sub_geocode={self.sub_code} hm={self.hectometer}"


@dataclass
class KmSubGeocode:
    object_id: int
    geo_code: str
    sub_code: str
    naam: str
    geometry: Polygon

    def xy_within(self, x: float, y: float):
        return Point(x, y).within(self.geometry)


@dataclass
class KmValueObject:
    geometry: Polygon
    km_vlak: KmVlak
    sub_geocode: KmSubGeocode
    km_punten: list[KmPunt] = field(default_factory=list)
    km_raaien: list[KmRaai] = field(default_factory=list)
    matched: list[KmPunt | KmRaai | None] = field(default_factory=list)

    def __str__(self):
        return f"{self.sub_geocode.geo_code}_{self.sub_geocode.sub_code}_{self.km_vlak.km_lint_naam} lint_naam={self.km_vlak.km_lint_naam} geo_code={self.sub_geocode.geo_code} sub_gecode={self.sub_geocode.sub_code} intersection={self.geometry.wkt}"

    def __repr__(self):
        self.__str__()



from __future__ import annotations

import numpy as np
import pyproj
from shapely.geometry import LineString, Point, Polygon


def gml_point_to_shapely(gml_point_coordinates: str) -> Point:
    """
    Converts a GML point coordinate string to a Shapely Point object.

    Args:
        gml_point_coordinates (str): The GML point coordinate string in the format "(x,y)".

    Returns:
        (Shapely.Point): The Shapely Point object.

    """
    coordinates = [
        float(x)
        for x in gml_point_coordinates.replace("(", "")
        .replace(")", "")
        .replace("'", "")
        .replace(" ", "")
        .split(",")
    ]
    return Point(coordinates)


def shapely_point_to_gml(shapely_point: Point):
    """
    Converts a Shapely Point object to a GML point coordinate string.

    Args:
        shapely_point (Point): The Shapely point".

    Returns:
        (str): gml:point string representation.

    """
    if shapely_point.has_z:
        return f"{round(shapely_point.x, 3)},{round(shapely_point.y, 3)},{round(shapely_point.z, 3)}"
    else:
        return f"{round(shapely_point.x, 3)},{round(shapely_point.y, 3)}"


def gml_linestring_to_shapely(gml_linestring_coordinates: str) -> LineString:
    """
    Converts a GML linestring coordinate string to a Shapely LineString object.

    Args:
        gml_linestring_coordinates (str): A string of GML linestring coordinates in "x,y" format separated by spaces.

    Returns:
        (Shapely.LineString): A Shapely LineString object.

    """
    return LineString(
        [tuple(map(float, x.split(","))) for x in gml_linestring_coordinates.split(" ")]
    )


def gml_polygon_to_shapely(gml_linestring_coordinates: str) -> Polygon:
    """
    Converts a GML polygon to a Shapely Polygon object.

    Args:
        gml_linestring_coordinates (str): A string containing the GML coordinates of the polygon.

    Returns:
        (Polygon): A Shapely Polygon object.

    """
    return Polygon(
        [tuple(map(float, x.split(","))) for x in gml_linestring_coordinates.split(" ")]
    )


class ShapelyTransform:
    """A utility class to transform between RD and WGS84 coordinate systems."""

    rd = pyproj.CRS("EPSG:28992")
    wgs = pyproj.CRS("EPSG:4326")
    transformer_to_wgs = pyproj.Transformer.from_crs(rd, wgs)
    transformer_to_rd = pyproj.Transformer.from_crs(wgs, rd)

    @classmethod
    def rd_to_wgs(
        cls, shapely: Point | LineString | Polygon
    ) -> Point | LineString | Polygon:
        """
        Convert a Shapely geometry from Dutch RD (Rijksdriehoekstelsel) coordinates (EPSG:28992) to WGS84 coordinates (EPSG:4326).

        Args:
            shapely (Union[Point, LineString, Polygon]): A Shapely geometry in Dutch RD coordinates.

        Returns:
            (Union[Point, LineString, Polygon]): A Shapely geometry in WGS84 coordinates.

        """
        return cls._convert(shapely, cls.transformer_to_wgs)

    @classmethod
    def wgs_to_rd(
        cls, shapely: Point | LineString | Polygon
    ) -> Point | LineString | Polygon:
        """
        Convert a Shapely geometry from WGS84 coordinates (EPSG:4326) coordinates (EPSG:28992) to Dutch RD (Rijksdriehoekstelsel).

        Args:
            shapely (Union[Point, LineString, Polygon]): A Shapely geometry in Dutch RD coordinates.

        Returns:
            (Union[Point, LineString, Polygon]): A Shapely geometry in WGS84 coordinates.

        """
        return cls._convert(shapely, cls.transformer_to_rd)

    @staticmethod
    def _convert(
        shapely: Point | LineString | Polygon, transformer: pyproj.Transformer
    ) -> Point | LineString | Polygon:
        if isinstance(shapely, Point):
            return Point(*reversed(transformer.transform(shapely.x, shapely.y)))

        elif isinstance(shapely, LineString):
            return LineString(zip(*reversed(transformer.transform(*shapely.coords.xy))))

        elif isinstance(shapely, Polygon):
            return LineString(
                zip(*reversed(transformer.transform(*shapely.exterior.coords.xy)))
            )
        else:
            return shapely


def reverse_line(shapely_polyline: LineString) -> LineString:
    """
    Reverses the order of coordinates in a Shapely LineString object.

    Args:
        shapely_polyline (LineString): The LineString object to reverse.

    Returns:
        (LineString): A new LineString object with the coordinates in reverse order.

    """
    return LineString(list(shapely_polyline.coords)[::-1])


def get_azimuth_from_points(point1: Point, point2: Point) -> float:
    """
    Calculates the azimuth angle between two points.

    Args:
        point1 (Point): The first Point object.
        point2 (Point): The second Point object.

    Returns:
        (float): The azimuth angle in degrees.

    """
    angle = np.arctan2(point2.x - point1.x, point2.y - point1.y)
    return float(np.degrees(angle)) if angle >= 0 else float(np.degrees(angle) + 360)


def check_point_in_area(point_str: str, area: Polygon):
    point_to_test = Point([float(item) for item in point_str.split(",")])
    return point_to_test.within(area)



import asyncio
from collections.abc import Callable
from copy import deepcopy

import numpy as np
from arcGisFeatureCache import ArcGisFeatureService
from shapely import LineString, MultiLineString, Polygon, STRtree

from kmService.km_models import KmPunt, KmRaai, KmSubGeocode, KmValueObject, KmVlak
from kmService.uitils.log import logger
from kmService.uitils.shapleyTools import extend_line


class KmServiceBuilder:
    """
    A builder class for constructing KmService instances.

    Attributes:
        value_objects: A dictionary containing key-value pairs of identifier and KmValueObject.
        url: The URL for the ArcGIS feature service.
        feature_service: An instance of ArcGisFeatureService.
    """

    def __init__(self):
        self.value_objects: dict[str, KmValueObject] = {}
        self._matched: dict[str, tuple[KmPunt, KmRaai]] = {}
        self.url: str = ""
        self.feature_service: ArcGisFeatureService | None = None
        self._km_vlakken: list[KmVlak] = []
        self._km_raaien: list[KmRaai] = []
        self._km_punten: list[KmPunt] = []
        self._sub_geocode: list[KmSubGeocode] = []

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_transport"]
        return state

    @classmethod
    async def factory(cls, url):
        """
        Factory method to create an instance of KmServiceBuilder.

        Args:
            url (str): The URL for the ArcGIS feature service.

        Returns:
            KmServiceBuilder: An instance of KmServiceBuilder.
        """
        self = KmServiceBuilder()
        self.url = url
        self.feature_service = await ArcGisFeatureService.factory(self.url)
        await self._process()
        return self

    async def _process(self):
        """
        Processes the ArcGIS feature service.
        """
        feature_info = {
            "Kilometerlintvlak": {
                "attributes": ["OBJECTID", "KMLINT", "KMLINT_OMSCHRIJVING"],
                "data_class": KmVlak,
                "data_list": self._km_vlakken,
            },
            "Raai": {
                "attributes": ["OBJECTID", "GEOCODE", "SUBCODE", "HECTOMETER"],
                "data_class": KmRaai,
                "data_list": self._km_raaien,
            },
            "Hectometerpunt (geocode)": {
                "attributes": [
                    "OBJECTID",
                    "GEOCODE",
                    "SUBCODE",
                    "KM_GEOCODE",
                    "KMLINT",
                    "KMLINT_OMSCHRIJVING",
                ],
                "data_class": KmPunt,
                "data_list": self._km_punten,
            },
            "Geocodesubgebied": {
                "attributes": ["OBJECTID", "GEOCODE", "SUBCODE", "NAAM"],
                "data_class": KmSubGeocode,
                "data_list": self._sub_geocode,
            },
        }

        tasks = []
        for feature_type in feature_info.keys():
            tasks.append(
                self._process_features(
                    feature_type,
                    feature_info[feature_type]["data_list"],
                    feature_info[feature_type]["data_class"],
                    feature_info[feature_type]["attributes"],
                )
            )
        await asyncio.gather(*tasks)

        await self._match_all_km_points_km_raai()

    async def _process_features(self, feature_type, data_list, data_class, attributes):
        """
        Processes the features of a specific type.

        Args:
            feature_type (str): The type of feature.
            data_list (list): The list to store the processed data.
            data_class (class): The class of the data.
            attributes (list): The attributes of the feature.
        """

        async def process_feature_item(item):
            _ = [item.attributes.get_value(attr) for attr in attributes]
            data_list.append(data_class(*_, geometry=item.geometry))

        if self.feature_service:
            logger.info(f"process {feature_type} layer")
            await asyncio.gather(
                *[
                    process_feature_item(item)
                    for item in self.feature_service.get_all_features([feature_type])
                ]
            )
        else:
            logger.error("no feature service")

    async def _create_value_object_sub_geocode_km_vlakken(
        self,
    ) -> dict[str, KmValueObject]:
        """
        Creates value objects for sub geocode and km vlakken.

        Returns:
            dict: A dictionary containing key-value pairs of identifier and KmValueObject.
        """

        async def builder(sub_geocode: KmSubGeocode, km_vlak: KmVlak):
            """
            Builds value objects for sub geocode and km vlakken.

            Args:
                sub_geocode (KmSubGeocode): The sub geocode object.
                km_vlak (KmVlak): The km vlak object.

            Returns:
                dict: A dictionary containing key-value pairs of identifier and KmValueObject.
            """
            _ = {}
            if sub_geocode.geometry.intersects(km_vlak.geometry):
                intersection = sub_geocode.geometry.intersection(km_vlak.geometry)
                tester = KmValueObject(intersection, km_vlak, sub_geocode)

                if isinstance(intersection, Polygon):
                    logger.success(f"created sub geocode km lint vlak {tester}")
                    _[
                        f"{sub_geocode.geo_code}_{sub_geocode.sub_code}_{km_vlak.km_lint_naam}"
                    ] = tester

                elif isinstance(intersection, LineString | MultiLineString):
                    logger.info(f"builder returns no polygon: {tester}")
            return _

        tasks = [
            builder(sub_geocode, km_vlak)
            for sub_geocode in self._sub_geocode
            for km_vlak in self._km_vlakken
        ]
        feature_data = await asyncio.gather(*tasks)
        out_dict = {k: v for d in feature_data for k, v in d.items()}

        return out_dict

    @staticmethod
    async def _match_km_punt_with_raai(km_punt_dict: dict, km_raai_dict: dict) -> dict:
        """
        Matches KmPunt objects with KmRaai objects.

        Args:
            km_punt_dict (dict): A dictionary containing KmPunt objects.
            km_raai_dict (dict): A dictionary containing KmRaai objects.

        Returns:
            dict: A dictionary containing matched KmPunt and KmRaai objects.
        """
        km_and_raai_match_dict = {}

        for key, value in km_punt_dict.items():
            # todo make async
            possible_match = []
            try:
                possible_match = km_raai_dict[
                    f"{value[0].geo_code}_{value[0].sub_code}"
                ]
            except KeyError:
                logger.warning(f"no raai for {value[0].geo_code}_{value[0].sub_code}")

            if possible_match is None:
                logger.warning(f"no raai for {value[0].geo_code}_{value[0].sub_code}")
            else:
                for item in value:
                    match_raai = [
                        item2
                        for item2 in possible_match
                        if item2.hectometer == item.hectometer
                    ]
                    assert len(match_raai) <= 1, "should be zero or one result"

                    key = f"{item.geo_code}_{item.sub_code}_{item.km_lint}"
                    if key not in km_and_raai_match_dict:
                        km_and_raai_match_dict[key] = [
                            [item, match_raai[0] if match_raai else None]
                        ]
                    else:
                        km_and_raai_match_dict[key].append(
                            [item, match_raai[0] if match_raai else None]
                        )

        return km_and_raai_match_dict

    @staticmethod
    async def _correct_km_raai(km_and_raai_match_dict: dict) -> None:
        """
        Corrects KmRaai objects.

        Args:
            km_and_raai_match_dict (dict): A dictionary containing matched KmPunt and KmRaai objects.
        """

        async def process_item(km, raai):
            if not raai:
                logger.warning(f"no raai for {km}")
            else:
                projected_point = raai.geometry_extended.interpolate(
                    raai.geometry_extended.project(km.geometry)
                )
                dx = km.geometry.x - projected_point.x
                dy = km.geometry.y - projected_point.y

                moved_line = LineString(
                    [(x + dx, y + dy) for x, y in raai.geometry_extended.coords]
                )

                distance = moved_line.distance(raai.geometry_extended)

                raai.geometry_corrected = moved_line
                logger.info(
                    f"km raai {raai} matched and moved to km punt, "
                    f"old distance to line: {round(distance, 3)}, "
                    f"new distance to line: {round(km.geometry.distance(moved_line), 3)}",
                    data={
                        "original_line": raai.geometry_extended,
                        "moved_line": moved_line,
                    },
                )

        tasks = []
        for key, value in km_and_raai_match_dict.items():
            tasks.extend([process_item(km, raai) for km, raai in value])
        await asyncio.gather(*tasks)

    @staticmethod
    async def _create_dict_from_list(items: list, key_function: Callable) -> dict:
        """
        Creates a dictionary from a list of items.

        Args:
            items (list): The list of items.
            key_function (Callable): A function to generate keys.

        Returns:
            dict: A dictionary containing the items.
        """
        result_dict: dict = {}
        for item in items:
            # todo make async
            key = key_function(item)
            if key in result_dict:
                result_dict[key].append(item)
            else:
                result_dict[key] = [item]
        return result_dict

    async def _match_all_km_points_km_raai(self):
        """
        Matches all KmPunt objects with KmRaai objects.
        """
        sub_geocode_km_vlakken = (
            await self._create_value_object_sub_geocode_km_vlakken()
        )
        logger.info("Create km punten and raaien dicts")
        km_punt_dict = await self._create_dict_from_list(
            self._km_punten,
            lambda km_punt: f"{km_punt.geo_code}_{km_punt.sub_code}_{km_punt.km_lint}",
        )
        km_raai_dict = await self._create_dict_from_list(
            self._km_raaien, lambda km_raai: f"{km_raai.geo_code}_{km_raai.sub_code}"
        )
        self._matched = await self._match_km_punt_with_raai(km_punt_dict, km_raai_dict)
        await self._correct_km_raai(self._matched)

        # add to value object
        for key, value in self._matched.items():
            if key in sub_geocode_km_vlakken:
                sub_geocode_km_vlakken[key].km_punten = [
                    p[0] for p in value if p[0] is not None
                ]
                sub_geocode_km_vlakken[key].km_raaien = [
                    p[1] for p in value if p[1] is not None
                ]
                sub_geocode_km_vlakken[key].matched = value
            else:
                raise ValueError("no split sub geocodes")  # noqa TRY003

        self.value_objects = sub_geocode_km_vlakken

        # check if no other raai is laying on the km_point, takes way to ling, use tree to get all objects with in 1m, filter on raai.
        temp = []
        temp_2 = {}
        for key, value in self.value_objects.items():
            for item in value.km_raaien:
                temp.append(item)
                temp_2[item.object_id] = item

        # make sure this is done before cutting raaien on sub geocode
        tree = STRtree([record.geometry_corrected for record in temp])
        tree_keys = np.array([record.object_id for record in temp])

        for key, value in self.value_objects.items():
            for item in value.matched:
                if not item[1]:
                    # todo: create tree of all raaien on corrected geometry
                    query = tree.query(
                        item[0].geometry, predicate="dwithin", distance=100
                    )
                    keys = tree_keys.take(query).tolist()
                    tester = [temp_2[item] for item in keys]
                    for _ in tester:
                        if _.hectometer == item[0].hectometer:
                            __ = deepcopy(_)
                            __.geometry_corrected = extend_line(__.geometry, 50)
                            item[1] = __

        # corrected cut on sub geocode polygon
        for key, value in self.value_objects.items():
            for item in value.km_raaien:
                tester = value.geometry.intersection(item.geometry_corrected)
                item.geometry_corrected = tester



import math
from dataclasses import dataclass, field

from shapely import LineString, Point

from kmService.km_models import KmValueObject
from kmService.uitils.shapely_geojson import (
    GeoJsonFeature,
    GeoJsonFeatureCollection,
    dumps,
)
from kmService.uitils.shapely_helpers import ShapelyTransform

# todo: make protocol for geojsononable objects


@dataclass
class GeoCodeResponse:
    """
    Represents geographical code response.

    Attributes:
        number: The geographical number.
        sub_code: The geographical sub-code.
        name: The geographical name.
    """

    number: int = -9999
    sub_code: str = ""
    name: str = ""


def km_string_helper(
    hm: float, distance: float, km_lint_name: str | None = None
) -> str:
    if distance:
        distance = math.floor(distance)
        if distance > 100:
            return f"km {hm:.3f} +{distance} {f'({km_lint_name})' if km_lint_name else ''})"
        else:
            if distance < 10:
                return f"km {hm:.1f}0{distance} ({km_lint_name})"
            else:
                return f"km {hm:.1f}{distance} ({km_lint_name})"
    else:
        return "no km"


@dataclass
class KmLintResponse:
    """
    Represents kilometer lint response.

    Attributes:
        name: The kilometer lint name.
        description: The description of the kilometer lint.
    """

    name: str = ""
    description: str = ""


@dataclass
class KmLintMeasure:
    """
    Represents a kilometer lint measure.

    Attributes:
        input_point (Point): The input point.
        hm (float): The hm value.
        distance (float): The distance.
        km_lint (KmLintResponse): The kilometer lint response.
        geocode (GeoCodeResponse): The geocode response.
        raai (LineString): The raai value.
    """

    input_point: Point = field(default_factory=Point)
    hm: float = float("nan")
    distance: float = float("nan")
    km_lint: KmLintResponse = field(default_factory=KmLintResponse)
    geocode: GeoCodeResponse = field(default_factory=GeoCodeResponse)
    raai: LineString = LineString()

    @classmethod
    def from_value_object(
        cls, km_value_object: KmValueObject, input_point: Point
    ) -> "KmLintMeasure":
        self = cls()
        self.input_point = input_point
        self.km_lint = KmLintResponse(
            km_value_object.km_vlak.km_lint_naam,
            km_value_object.km_vlak.km_lint_omschrijving,
        )
        self.geocode = GeoCodeResponse(
            int(km_value_object.sub_geocode.geo_code),
            km_value_object.sub_geocode.sub_code,
            km_value_object.sub_geocode.naam,
        )
        return self

    @property
    def display(self) -> str:
        """
        Computes the kilometer value string.
        """
        return km_string_helper(self.distance, self.hm, self.km_lint.name)

    def get_geojson_features(self) -> list[GeoJsonFeature]:
        # """
        # Gets the GeoJSON features.
        #
        # Returns:
        #     The list of GeoJSON features.
        # """
        features = [
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(self.raai)], {"type": "raai", "hm": self.hm}
            ),
            # todo: make point on line and create line between input
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(self.input_point)],
                {
                    "km_value": self.display,
                    "hm": self.hm,
                    "distance": self.distance,
                    "km_lint_name": self.km_lint.name,
                    "km_lint_description": self.km_lint.description,
                    "geocode_number": self.geocode.number,
                    "sub_geocode": self.geocode.sub_code,
                    "geocode_name": self.geocode.name,
                },
            ),
        ]
        return features

    def get_geojson_feature_collection(self) -> GeoJsonFeatureCollection:
        """
        Retrieve a GeoJSON FeatureCollection object.

        Returns:
            A collection of GeoJSON features.
        """
        return GeoJsonFeatureCollection(self.get_geojson_features())

    def geojson_string(self) -> str:
        """
        Gets the GeoJSON string representation.

        Returns:
            The GeoJSON string.
        """
        return dumps(self.get_geojson_feature_collection(), indent=4)


@dataclass
class KmResponse:
    """
    Represents a response containing kilometer measures.

    Attributes:
        input_point (Point): The input point associated with the response.
        km_measures (List[KmLintMeasure]): A list of kilometer measures.
    """

    input_point: Point = Point()
    km_measures: list[KmLintMeasure] = field(default_factory=list)

    @property
    def has_km_measure(self) -> bool:
        """
        Check if the response contains at least one kilometer measure.
        """
        if len(self.km_measures) != 0:
            return True
        return False

    @property
    def multiple_km_measure(self) -> bool:
        """
        Check if the response contains multiple kilometer measures.
        """
        if len(self.km_measures) != 1:
            return True
        return False

    @property
    def display(self) -> str:
        """
        Computes the kilometer value string, if multiple response returns one string separated by a ', '.
        """
        if self.multiple_km_measure:
            return ", ".join(
                [
                    km_string_helper(_.hm, _.distance, _.km_lint.name)
                    for _ in self.km_measures
                ]
            )

        return [
            km_string_helper(_.hm, _.distance, _.km_lint.name) for _ in self.km_measures
        ][0]

    def get_geojson_features(self) -> list[GeoJsonFeature]:
        # """
        # Generate a list of GeoJSON features from the kilometer measures.
        #
        # Returns:
        #     A list of GeoJSON features representing the kilometer measures.
        # """
        geojson_features = []
        for item in self.km_measures:
            geojson_features.extend(item.get_geojson_features())
        return geojson_features

    def get_geojson_feature_collection(self) -> GeoJsonFeatureCollection:
        """
        Retrieve a GeoJSON FeatureCollection object.

        Returns:
            A collection of GeoJSON features.
        """
        return GeoJsonFeatureCollection(self.get_geojson_features())

    def geojson_string(self) -> str:
        """
        Generate GeoJSON representation of the response.

        Returns:
            GeoJSON string representing the kilometer measures.
        """
        return dumps(self.get_geojson_feature_collection(), indent=4)


@dataclass
class ProjectionInputResponse:
    """
    Represents the input response for a point.

    !!! warning "This feature is experimental and may not be suitable for production use."
        ***Api and response models will change!***
        Ensure to thoroughly test this feature with your data before deploying it.

    Args:
        input_line: The input line.
        input_lint_name: The input line name.
        input_hm: The input kilometers.
        input_distance: The input distance.

    """

    input_line: LineString
    input_lint_name: str
    input_hm: float
    input_distance: float


@dataclass
class ProjectionResponse:
    """
    Represents the response for a point.

    !!! warning "This feature is experimental and may not be suitable for production use."
        ***Api and response models will change!***
        Ensure to thoroughly test this feature with your data before deploying it.

    Args:
        input_data: The input data for the response.
        km_point: The projected kilometering point.
        km_line: The kilometering line (offset on distance of input raai).
        reference_line: The reference line, optional.
    """

    input_data: ProjectionInputResponse
    km_line: LineString
    km_point: Point
    reference_line: LineString | None = None

    def get_geojson_features(self) -> list[GeoJsonFeature]:
        features = [
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(self.input_data.input_line)],
                {"type": "input_line_geometry", "input_data": self.input_data.__dict__},
            ),
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(self.km_line)],
                {
                    "type": "km_line",
                    "km": self.input_data.input_hm,
                    "lint": self.input_data.input_lint_name,
                },
            ),
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(self.km_point)], {"type": "projected_point"}
            ),
        ]
        if self.reference_line:
            features.append(
                GeoJsonFeature(
                    [ShapelyTransform.rd_to_wgs(self.reference_line)],
                    {"type": "raai", "km": self.input_data.input_hm},
                )
            )
        return features

    def get_geojson_feature_collection(self) -> GeoJsonFeatureCollection:
        """
        Retrieve a GeoJSON FeatureCollection object.

        Returns:
            A collection of GeoJSON features.
        """
        return GeoJsonFeatureCollection(self.get_geojson_features())

    def geojson_string(self) -> str:
        """
        Converts the response to GeoJSON format.

        Returns:
            The GeoJSON representation of the response.
        """
        return dumps(self.get_geojson_feature_collection(), indent=4)



from .km_responses import (
    GeoCodeResponse,
    KmLintMeasure,
    KmLintResponse,
    KmResponse,
    ProjectionInputResponse,
    ProjectionResponse,
)
from .km_service import KmService, KmServiceBuilder, get_km_service

__version__ = "0.0.1a1"

__all__ = [
    "KmService",
    "KmServiceBuilder",
    "get_km_service",
    "GeoCodeResponse",
    "KmLintResponse",
    "KmLintMeasure",
    "KmResponse",
    "ProjectionInputResponse",
    "ProjectionResponse",
]



from loguru import logger

# possible log levels:
#   - TRACE (5): used to record fine-grained information about the program's execution path for diagnostic purposes.
#   - DEBUG (10): used by developers to record messages for debugging purposes.
#   - INFO (20): used to record informational messages that describe the normal operation of the program.
#   - SUCCESS (25): similar to INFO but used to indicate the success of an operation.
#   - WARNING (30): used to indicate an unusual event that may require further investigation.
#   - ERROR (40): used to record error conditions that affected a specific operation.
#   - CRITICAL (50): used to record error conditions that prevent a core function from working.

logger.add("log.log", level=0, catch=True, backtrace=True, diagnose=True)
logger.opt(record=True)
logger.level = 0  # type: ignore


# jupyter notebook logging massive and slow logging solution: disable logging, if need log print it.


def run_from_notebook():  # pragma: no cover
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


# remove logger if jupyter notebook
if run_from_notebook():  # pragma: no cover
    try:
        logger.remove(0)
    except ValueError:
        pass


# hacky magic print methode
def print_log(from_row=0):  # pragma: no cover
    log_sink_path = logger._core.handlers[1]._sink._file_path
    with open(log_sink_path) as f:
        x = f.readlines()
    print("".join(x[from_row:]))



"""
library with functions to edit and perform spatial analysis on vector data.
build around shapely and fiona libraries

Created on 2016-July-16
@author: Dirk Eilander (dirk.eilander@deltares.nl)

"""

import numpy as np
import rtree
import shapely.ops
from shapely.geometry import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)


def snap_lines(lines, max_dist, tolerance=1e-3, return_index=False):
    """Snap lines together if the endpoint of one line1 is at most max_dist apart from the line2.
    Both lines are snapped by extending line1 towards line2
    the max distance is measured in the direction of the line1

    Args:
        lines: a list of LineStrings or a MultiLineString
        max_dist: maximum distance two endpoints may be joined together

    """
    # extend all lines with max_dist to snap lines at a sharp angle
    lines_ext = [extend_line(line, max_dist) for line in lines]

    # build spatial index of line bounding boxes
    tree_idx = rtree.index.Index()
    lines_bbox = [_.bounds for _ in lines_ext]
    for i, bbox in enumerate(lines_bbox):
        tree_idx.insert(i, bbox)

    lines_snap = []
    idx_snapped = []
    for i1, line in enumerate(lines):
        if isinstance(line, LineString):
            for side in ["start", "end"]:
                coords = line.coords[:]
                # make line extensions
                if side == "start":
                    ext = LineString(
                        [shift_point(coords[0], coords[1], -1.0 * max_dist), coords[0]]
                    )
                    pnt_from = Point(coords[0])
                elif side == "end":
                    ext = LineString(
                        [
                            coords[-1],
                            shift_point(coords[-1], coords[-2], -1.0 * max_dist),
                        ]
                    )
                    pnt_from = Point(coords[-1])

                # point instead of line if max_dist is zero (only clipping)
                if max_dist == 0:
                    ext = pnt_from

                # find close-by lines based on bounds with spatial index
                if tolerance > 0:
                    hits = list(tree_idx.intersection(ext.buffer(tolerance).bounds))
                else:
                    hits = list(tree_idx.intersection(ext.bounds))
                lines_hit = MultiLineString([lines_ext[i] for i in hits if i != i1])
                # find intersection points. function yields list of points
                int_points = intersection_points([ext], lines_hit, tolerance=tolerance)
                if len(int_points) == 0:
                    continue

                # if intersection yields something else then a Point, break down to nearest point
                if len(int_points) > 1:
                    # if more intersections, find closest
                    pnt_to = closest_object(int_points, pnt_from)[0]
                else:
                    pnt_to = int_points[0]

                # at this point pnt_to is the closest intersecting point
                if pnt_from != pnt_to:  # check if lines are not already touching
                    # snap line towards pnt_to
                    if side == "start":
                        coords[0] = pnt_to.coords[0]
                    elif side == "end":
                        coords[-1] = pnt_to.coords[0]
                    line = LineString(coords)
                    if i1 not in idx_snapped:
                        # bookkeeping: list with changes features
                        idx_snapped.append(i1)

            lines_snap.append(line)

    if return_index:
        return lines_snap, idx_snapped
    else:
        return lines_snap


# split methods
def split_lines(lines, points=None, tolerance=0.0, return_index=False):
    """split lines at intersection, or if given, at points"""
    if isinstance(points, Point):
        points = [points]

    # create output list
    lines_out = []
    index_out = []

    # build spatial index
    if points is None:
        # build spatial index of line bounding boxes
        tree_idx = rtree.index.Index()
        lines_bbox = [_.bounds for _ in lines]
        for i, bbox in enumerate(lines_bbox):
            tree_idx.insert(i, bbox)
    else:
        # build spatial index of points
        tree_idx = rtree.index.Index()
        for i, p in enumerate(points):
            tree_idx.insert(i, p.buffer(tolerance).bounds)

    # loop through lines and split lines with split point
    for idx, line in enumerate(lines):
        if points is None:
            # find close-by lines based on bounds with spatial index
            hits = list(tree_idx.intersection(lines_bbox[idx]))
            lines_hit = MultiLineString([lines[i] for i in hits if i != idx])
            # find line intersections
            # lines_other = [l for i, l in enumerate(lines) if i != idx]
            split_points = intersection_points([line], lines_hit, tolerance=tolerance)
        else:
            # find close-by points based on bounds of line with spatial index
            hits = list(tree_idx.intersection(line.bounds))
            points_hit = [points[i] for i in hits]
            # find points which intersects with line
            split_points = [p for p in points_hit if line.distance(p) <= tolerance]
            # split_points = [p for p in points if line.distance(p) <= tolerance]

        # check if intersections for line
        if len(split_points) >= 1:
            for p in split_points:
                line = split_line(line, p, tolerance=tolerance)
            for _ in line:
                lines_out.append(_)
                index_out.append(idx)
        else:
            lines_out.append(line)
            index_out.append(idx)

    if return_index:
        return remove_redundant_nodes(lines_out), index_out
    else:
        return remove_redundant_nodes(lines_out)


def split_line(line, point, tolerance=0.0):
    """split line (Shapely LineString or MultiLineString) at  point (Shapely Point),
    return splitted line (Shapely MultiLineString)"""
    if not isinstance(line, LineString | MultiLineString):
        raise TypeError("line should be shapely LineString or MultiLineString object")  # noqa TRY003
    if not isinstance(point, Point):
        raise TypeError("point should be shapely Point object")  # noqa TRY003

    # function works with MultiLineStrings to be able to use the function in a split loop
    if not isinstance(line, MultiLineString):
        line = MultiLineString([line])
    lines_out = []

    # for intersecting line, find intersecting segment and split line
    for l0 in line:
        # check if point on line, but not one of its endpoints
        if (not point.touches(l0.boundary)) and (point.distance(l0) <= tolerance):
            coords = list(l0.coords)
            segments = [LineString(s) for s in pairs(coords)]
            for i, segment in enumerate(segments):
                # find intersecting segment
                if segment.distance(point) <= tolerance:
                    if Point(coords[i]).touches(point):
                        # split line at vertex if within tolerance
                        la = LineString(coords[: i + 1])
                        lb = LineString(coords[i:])
                        if (la.length > tolerance) & (lb.length > tolerance):
                            lines_out.append(la)
                            lines_out.append(lb)
                        else:
                            lines_out.append(l0)
                        break
                    else:
                        # split line at point on segment
                        la = LineString(coords[: i + 1] + [(point.x, point.y)])
                        lb = LineString([(point.x, point.y)] + coords[i + 1 :])
                        if (la.length > tolerance) & (lb.length > tolerance):
                            lines_out.append(la)
                            lines_out.append(lb)
                        else:
                            lines_out.append(l0)
                        break
        else:
            lines_out.append(l0)

    return MultiLineString(lines_out)


def explode_polygons(polygons, return_index=False):
    """returns main line features that make up the polygons"""
    lines_out = []
    index = []

    if isinstance(polygons, Polygon):
        polygons = [polygons]
    for i, line in enumerate(polygons):
        for linestring in [LineString(s) for s in pairs(line.exterior.coords)]:
            lines_out.append(linestring)
            index.append(i)
        for p in line.interiors:
            lines_out.append(LineString(p.coords))
            index.append(i)

    if not return_index:
        return lines_out
    else:
        return lines_out, index


def clip_lines_with_polygon(
    lines, polygon, tolerance=1e-3, within=True, return_index=False
):
    """clip lines based on polygon outline"""
    # get boundaries of polygon
    boundaries = explode_polygons(polygon)
    # find intersection points of boundaries and lines and split lines based on it
    int_points = intersection_points(lines, boundaries)
    lines_split, index = split_lines(
        lines, int_points, tolerance=tolerance, return_index=True
    )

    # select lines that are contained by polygon
    polygon_buffer = polygon.buffer(
        tolerance
    )  # small buffer to allow for use 'within' function
    if within:
        lines_clip = [line for line in lines_split if line.within(polygon_buffer)]
        index = [
            i for i, line in zip(index, lines_split) if line.within(polygon_buffer)
        ]
    else:
        lines_clip = [line for line in lines_split if not line.within(polygon_buffer)]
        index = [
            i for i, line in zip(index, lines_split) if not line.within(polygon_buffer)
        ]
    if not return_index:
        return lines_clip
    else:
        return lines_clip, index


# neighborhood methods
def closest_object(geometries, point):
    """Find the nearest geometry among a list, measured from fixed point.

    Args:
        geometries: a list of shapely geometry objects
        point: a shapely Point

    Returns:
        Tuple (geom, min_dist, min_index) of the geometry with minimum distance
        to point, its distance min_dist and the list index of geom, so that
        geom = geometries[min_index].
    """
    min_dist, min_index = min(
        (point.distance(geom), k) for (k, geom) in enumerate(geometries)
    )

    return geometries[min_index], min_dist, min_index


def intersection_points(lines1, lines2=None, tolerance=0.0, min_spacing=0):
    """creates list with points of line intersections. if intersection is other type than a point,
    it is broken down to points of its coordinates

    :param lines1: MultiLineString or list of lines
    :param lines2: MultiLineString or list of lines, if None find intersections amongst lines1
    :return:        list with shapely points of intersection
    """
    points = []
    tree_idx_pnt = rtree.index.Index()
    ipnt = 0

    if lines2 is None:
        # build spatial index for lines1
        tree_idx = rtree.index.Index()
        lines_bbox = [_.bounds for _ in lines1]
        for i, bbox in enumerate(lines_bbox):
            tree_idx.insert(i, bbox)

    # create multilinestring of close-by lines
    for i1, l1 in enumerate(lines1):
        if lines2 is None:
            # find close-by lines based on bounds with spatial index
            hits = list(tree_idx.intersection(lines_bbox[i1]))
            lines_hit = MultiLineString([lines1[i] for i in hits if i != i1])
        else:
            lines_hit = MultiLineString(lines2)

        if tolerance > 0:
            l1 = extend_line(l1, tolerance)

        x = l1.intersection(lines_hit)
        if not x.is_empty:
            if isinstance(x, Point):
                pnts = [x]

            else:
                if isinstance(x, MultiPoint):
                    pnts = [Point(geom) for geom in x]
                elif isinstance(x, MultiLineString | MultiPolygon | GeometryCollection):
                    pnts = [Point(coords) for geom in x for coords in geom.coords]
                elif isinstance(x, LineString | Polygon):
                    pnts = [Point(coords) for coords in x.coords]
                else:
                    raise NotImplementedError("intersection yields bad type")

            for pnt in pnts:
                if min_spacing > 0:
                    if ipnt > 0:
                        hits = list(tree_idx_pnt.intersection(pnt.bounds))
                    else:
                        hits = []

                    if len(hits) == 0:  # no pnts within spacing
                        ipnt += 1
                        tree_idx_pnt.insert(ipnt, pnt.buffer(min_spacing).bounds)
                        points.append(pnt)
                else:
                    points.append(pnt)

    return points


# utils
def cap_lines(line, offset=0.0, length=1.0):
    """
    Prepare two cap lines at the beginning and end of a LineString object
    The function can be used to prepare boundary condition lines for a hydraulic model
    Args:
        line: LineString object
        offset=0.: float, determining an offset from the beginning/end point
        length=1.: length of the cap lines (in same units as LineString coordinates)
    Returns:
        a list with two LineString objects, containing the cap lines
    """
    coords = line.coords
    if offset > 0:
        # get start & end point line
        start, end = Point(coords[0]), Point(coords[-1])
    else:  # create short line around endpoints (necessary for perpendicular line function)
        offset = 0.1
        # get start & end point line
        start = shift_point(coords[0], coords[1], 2 * abs(offset))
        end = shift_point(coords[-1], coords[-2], 2 * abs(offset))

    # extend line at both sides
    start_extend = shift_point(coords[0], coords[1], -2 * abs(offset))
    end_extend = shift_point(coords[-1], coords[-2], -2 * abs(offset))

    # make new line segments
    start_line = shapely.geometry.LineString([start, start_extend])
    end_line = shapely.geometry.LineString([end, end_extend])

    # get perpendicular lines at half length of start &  end lines
    cap1 = perpendicular_line(start_line, length)
    cap2 = perpendicular_line(end_line, length)
    return [cap1, cap2]


def shift_point(c1, c2, offset):
    """

    shift points with offset in orientation of line c1->c2
    """
    try:
        x1, y1 = c1
        x2, y2 = c2
    except Exception as e:
        print(e, c1, c2)
    if ((x1 - x2) == 0) and ((y1 - y2) == 0):  # zero length line
        x_new, y_new = x1, y1
    else:
        rel_length = np.minimum(offset / np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2), 1)
        x_new = x1 + (x2 - x1) * rel_length
        y_new = y1 + (y2 - y1) * rel_length
    return Point(x_new, y_new)


def extend_line(line, offset, side="both"):
    """extend line in same orientation"""
    if side == "both":
        sides = ["start", "end"]
    else:
        sides = [side]

    for side in sides:
        coords = line.coords
        if side == "start":
            p_new = shift_point(coords[0], coords[1], -1.0 * offset)
            line = LineString([p_new] + coords[:])
        elif side == "end":
            p_new = shift_point(coords[-1], coords[-2], -1.0 * offset)
            line = LineString(coords[:] + [p_new])
    return line


def perpendicular_line(l1, length):
    """Create a new Line perpendicular to this linear entity which passes
    through the point `p`.


    """
    dx = l1.coords[1][0] - l1.coords[0][0]
    dy = l1.coords[1][1] - l1.coords[0][1]

    p = Point(l1.coords[0][0] + 0.5 * dx, l1.coords[0][1] + 0.5 * dy)
    x, y = p.coords[0][0], p.coords[0][1]

    if (dy == 0) or (dx == 0):
        a = length / l1.length
        l2 = LineString(
            [(x - 0.5 * a * dy, y - 0.5 * a * dx), (x + 0.5 * a * dy, y + 0.5 * a * dx)]
        )

    else:
        s = -dx / dy
        a = ((length * 0.5) ** 2 / (1 + s**2)) ** 0.5
        l2 = LineString([(x + a, y + s * a), (x - a, y - s * a)])

    return l2


def pairs(lst):
    """Iterate over a list in overlapping pairs.

    Args:
        lst: an iterable/list

    Returns:
        Yields a pair of consecutive elements (lst[k], lst[k+1]) of lst. Last
        call yields (lst[-2], lst[-1]).

    Example:
        lst = [4, 7, 11, 2]
        pairs(lst) yields (4, 7), (7, 11), (11, 2)

    Source:
        http://stackoverflow.com/questions/1257413/1257446#1257446
    """
    i = iter(lst)
    prev = next(i)
    for item in i:
        yield prev, item
        prev = item


# coordinates of geoms
def add_coordinate(line, distance):
    # Cuts a line in two at a distance from its starting point
    if distance <= 0.0 or distance >= line.length:
        return line
    coords = list(line.coords)
    for i, p in enumerate(coords):
        pd = line.project(Point(p))
        if pd == distance:
            break
        if pd > distance:
            cp = line.interpolate(distance)
            line = LineString(coords[:i] + [(cp.x, cp.y)] + coords[i:])
            break
    return line


def increase_points_line(line, spacing):
    line_length = line.length
    cp = Point(line.interpolate(line_length - spacing))
    line = LineString(line.coords[:-1] + [(cp.x, cp.y)] + [line.coords[-1]])
    for i, d in enumerate(np.arange(line_length, spacing, -spacing)):
        line = add_coordinate(line, d)
    return line


def remove_redundant_nodes(lines, tolerance=1e-7):
    """remove vertices with length smaller than tolerance"""
    lines_out = []
    for line in lines:
        coords = line.coords
        l_segments = np.array(
            [Point(s[0]).distance(Point(s[1])) for s in pairs(coords)]
        )
        idx = np.where(l_segments < tolerance)[0]
        lines_out.append(LineString([c for i, c in enumerate(coords) if i not in idx]))
    return lines_out
